# S3 Immutable Bucket Cleanup Tool - Simplified Version
# This script removes objects from S3 buckets where the current version
# is a delete marker, respecting Object Lock retention policies.

import datetime
import logging
import sys
import os
import json
import configparser
from concurrent.futures import ThreadPoolExecutor, as_completed
from collections import defaultdict
import glob


# --- START: Add local 'vendor' directory to path ---
# This allows the script to find and import the boto3 library without installation.
try:
    # Get the absolute path of the directory where the script is located
    script_dir = os.path.dirname(os.path.abspath(__file__))
    # Define the path to the vendor directory
    vendor_dir = os.path.join(script_dir, 'vendor')
    # Add the vendor directory to the system path
    if vendor_dir not in sys.path:
        sys.path.insert(0, vendor_dir)
except NameError:
    # Fallback for environments where __file__ is not defined
    vendor_dir = os.path.join(os.getcwd(), 'vendor')
    if vendor_dir not in sys.path:
        sys.path.insert(0, vendor_dir)
# --- END: Add local 'vendor' directory to path ---
import boto3

# --- CONFIGURATION ---
RETENTION_CHECK_THREADS = 2
BULK_DELETE_SIZE = 300

try:
    # Get the directory where this script file resides
    script_dir = os.path.dirname(os.path.abspath(__file__))
except NameError:
    # Fallback if __file__ isn’t defined (e.g., some interactive runners)
    script_dir = os.getcwd()

# Change the working directory to the script’s directory
os.chdir(script_dir)


def load_config():
    """Load configuration from config.ini file"""
    config = configparser.ConfigParser()

    if not os.path.exists('config.ini'):
        print("Error: config.ini file not found")
        sys.exit(1)

    try:
        config.read('config.ini')

        # Validate required section exists
        if 'S3' not in config:
            print("Error: [S3] section not found in config.ini")
            sys.exit(1)

        s3_config = config['S3']

        # Extract and validate required values
        endpoint_url = s3_config.get('ENDPOINT_URL', '').strip()
        access_key = s3_config.get('ACCESS_KEY', '').strip()
        secret_key = s3_config.get('SECRET_KEY', '').strip()
        bucket_name = s3_config.get('BUCKET_NAME', '').strip()

        # Validate required fields
        if not endpoint_url:
            print("Error: ENDPOINT_URL is required in config.ini")
            sys.exit(1)

        if not bucket_name:
            print("Error: BUCKET_NAME is required in config.ini")
            sys.exit(1)

        # Validate endpoint URL format
        if not validate_endpoint_url(endpoint_url):
            print("Error: Invalid ENDPOINT_URL format in config.ini")
            print("Expected format: https://region-name-N.storage.impossibleapi.net")
            print(f"Configured URL:  {endpoint_url}")
            print("e.g., https://eu-central-2.storage.impossibleapi.net")
            sys.exit(1)

        return {
            'ENDPOINT_URL': endpoint_url,
            'ACCESS_KEY': access_key,
            'SECRET_KEY': secret_key,
            'BUCKET_NAME': bucket_name
        }

    except Exception as e:
        print("Error reading config.ini: {}".format(e))
        sys.exit(1)


def validate_endpoint_url(url):
    """Validate endpoint URL format"""
    if not url:
        return False

    # Check if URL starts with https://
    if not url.startswith('https://'):
        return False

    # Remove https:// prefix
    url_without_protocol = url[8:]  # len('https://') = 8

    # Remove trailing slash if present
    if url_without_protocol.endswith('/'):
        url_without_protocol = url_without_protocol[:-1]

    # Check if URL ends with storage.impossibleapi.net
    if not url_without_protocol.endswith('.storage.impossibleapi.net'):
        return False

    # Extract the region part (everything before .storage.impossibleapi.net)
    region_part = url_without_protocol[:-len('.storage.impossibleapi.net')]

    # Check if region part is not empty
    if not region_part:
        return False

    # Count dashes in region part
    dash_count = region_part.count('-')

    # Region must contain exactly 2 dashes
    if dash_count != 2:
        return False

    # Check that region doesn't start or end with dash
    if region_part.startswith('-') or region_part.endswith('-'):
        return False

    # Check that region contains only alphanumeric characters and dashes
    for char in region_part:
        if not (char.isalnum() or char == '-'):
            return False

    # Check that region doesn't have consecutive dashes
    if '--' in region_part:
        return False

    return True


def setup_logger(log_file_name):
    """Set up logger for INFO level only"""
    log_dir = "log"
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    log_file_path = os.path.join(log_dir, log_file_name)

    logger = logging.getLogger('S3BucketCleanup')
    logger.setLevel(logging.INFO)
    logger.handlers = []

    formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')

    # File handler
    fh = logging.FileHandler(log_file_path, mode='a')
    fh.setLevel(logging.INFO)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    # Console handler
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    return logger


def save_delete_batch_to_file(delete_batch, batch_number, bucket_name, logger):
    """Save delete batch to JSON file"""
    delete_objects_dir = "delete-objects"
    if not os.path.exists(delete_objects_dir):
        os.makedirs(delete_objects_dir)

    filename = os.path.join(delete_objects_dir,
                            "{}-delete-objects{:03d}.json".format(bucket_name, batch_number))

    clean_delete_batch = []
    for obj in delete_batch:
        clean_delete_batch.append({
            'Key': obj['Key'],
            'VersionId': obj['VersionId']
        })

    delete_request = {
        "Delete": {
            "Objects": clean_delete_batch,
            "Quiet": False
        }
    }

    try:
        with open(filename, 'w') as f:
            json.dump(delete_request, f, indent=2, default=str)
        logger.info("Saved delete batch to: %s (%d objects)", filename, len(clean_delete_batch))
        return filename
    except Exception as e:
        logger.error("Failed to save delete batch to %s: %s", filename, e)
        return None


def execute_bulk_delete(s3_client, bucket_name, delete_batch, logger):
    """Execute bulk delete operation"""
    if not delete_batch:
        logger.info("Empty delete batch, skipping bulk delete operation")
        return 0, 0

    try:
        clean_delete_batch = []
        for obj in delete_batch:
            clean_delete_batch.append({
                'Key': obj['Key'],
                'VersionId': obj['VersionId']
            })

        response = s3_client.delete_objects(
            Bucket=bucket_name,
            Delete={
                'Objects': clean_delete_batch,
                'Quiet': False
            }
        )

        deleted = response.get('Deleted', [])
        errors = response.get('Errors', [])

        success_count = len(deleted)
        error_count = len(errors)

        # Log errors
        for error in errors:
            error_code = error.get('Code', '')
            error_message = error.get('Message', '')
            key = error.get('Key', '')
            version_id = error.get('VersionId', '')
            logger.error("Failed to delete %s (version %s): %s - %s",
                         key, version_id, error_code, error_message)

        return success_count, error_count

    except Exception as e:
        logger.error("Bulk delete operation failed: %s", e)
        return 0, len(delete_batch)


def check_object_has_retention(s3_client, bucket_name, key, versions, logger):
    """Check if any version of an object has active retention protection"""
    try:
        for version in versions:
            if version['IsDeleteMarker']:
                continue

            try:
                response = s3_client.get_object_retention(
                    Bucket=bucket_name,
                    Key=key,
                    VersionId=version['VersionId']
                )

                retention = response.get('Retention', {})
                if retention:
                    retain_until = retention.get('RetainUntilDate')
                    if retain_until:
                        now = datetime.datetime.now(datetime.timezone.utc)
                        if now < retain_until:
                            return True

            except Exception as e:
                error_code = None
                if hasattr(e, 'response') and 'Error' in e.response:
                    error_code = e.response['Error'].get('Code')

                if error_code in ['NoSuchObjectLockRetention',
                                  'ObjectLockConfigurationNotFoundError']:
                    continue
                else:
                    return True

        return False

    except Exception as e:
        logger.error("Error checking object retention for %s: %s", key, e)
        return True


def check_single_object_retention(args):
    """Check retention for a single object (for parallel processing)"""
    s3_client, bucket_name, key, versions, logger = args
    has_retention = check_object_has_retention(s3_client, bucket_name, key, versions, logger)
    return key, has_retention


def create_s3_client_for_retention_check(endpoint_url, access_key, secret_key):
    """Create a new S3 client for retention checking threads"""
    s3_client_args = {'endpoint_url': endpoint_url}
    if access_key:
        s3_client_args['aws_access_key_id'] = access_key
    if secret_key:
        s3_client_args['aws_secret_access_key'] = secret_key
    return boto3.client('s3', **s3_client_args)


def list_all_object_versions(s3_client, bucket_name, logger):
    """List all object versions in the bucket and group by key"""
    object_versions_map = defaultdict(list)

    try:
        paginator = s3_client.get_paginator('list_object_versions')
        page_number = 0

        logger.info("Starting to list all object versions...")

        for page in paginator.paginate(Bucket=bucket_name):
            page_number += 1
            logger.info("Processing page %d...", page_number)

            # Process regular versions
            for version_data in page.get('Versions', []):
                key = version_data['Key']
                object_versions_map[key].append({
                    'VersionId': version_data['VersionId'],
                    'Size': version_data.get('Size', 0),
                    'IsLatest': version_data['IsLatest'],
                    'LastModified': version_data['LastModified'],
                    'IsDeleteMarker': False
                })

            # Process delete markers
            for marker_data in page.get('DeleteMarkers', []):
                key = marker_data['Key']
                object_versions_map[key].append({
                    'VersionId': marker_data['VersionId'],
                    'Size': 0,
                    'IsLatest': marker_data['IsLatest'],
                    'LastModified': marker_data['LastModified'],
                    'IsDeleteMarker': True
                })

        logger.info("Finished listing versions. Processed %d page(s).", page_number)
        return object_versions_map

    except Exception as e:
        logger.error("Error listing object versions: %s", e)
        raise


def human_readable_size(size_bytes):
    """Convert bytes to human readable format"""
    if size_bytes == 0:
        return "0 B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB")
    i = 0
    size_bytes = float(size_bytes)
    while size_bytes >= 1024.0 and i < len(size_name) - 1:
        size_bytes /= 1024.0
        i += 1
    return "{:.2f} {}".format(size_bytes, size_name[i])


def cleanup_old_logs(log_dir="log", max_logs=10):
    """Keep only the most recent log files, delete older ones"""
    try:
        if not os.path.exists(log_dir):
            return

        # Find all log files matching the pattern
        log_pattern = os.path.join(log_dir, "s3_bucket_cleanup_*.log")
        log_files = glob.glob(log_pattern)

        if len(log_files) <= max_logs:
            return

        # Sort by modification time (newest first)
        log_files.sort(key=lambda x: os.path.getmtime(x), reverse=True)

        # Keep only the most recent max_logs files
        files_to_delete = log_files[max_logs:]

        for file_path in files_to_delete:
            try:
                os.remove(file_path)
                print("Deleted old log file: {}".format(os.path.basename(file_path)))
            except Exception as e:
                print("Warning: Could not delete log file {}: {}".format(file_path, e))

    except Exception as e:
        print("Warning: Error during log cleanup: {}".format(e))


def check_bucket_versioning(s3_client, bucket_name, logger):
    """Check if bucket has versioning enabled"""
    try:
        response = s3_client.get_bucket_versioning(Bucket=bucket_name)
        status = response.get('Status', 'Disabled')

        if status != 'Enabled':
            logger.error("Bucket versioning is not enabled. Current status: %s", status)
            logger.error("This tool only works with versioned buckets.")
            return False

        logger.info("Bucket versioning is enabled")
        return True
    except Exception as e:
        logger.error("Failed to check bucket versioning: %s", e)
        return False


def check_bucket_object_lock(s3_client, bucket_name, logger):
    """Check if bucket has Object Lock configuration"""
    try:
        response = s3_client.get_object_lock_configuration(Bucket=bucket_name)
        object_lock_enabled = response.get('ObjectLockConfiguration', {}).get('ObjectLockEnabled')

        if object_lock_enabled != 'Enabled':
            logger.error("Object Lock is not enabled on the bucket")
            logger.error("This tool is designed for Object Lock enabled buckets.")
            return False

        logger.info("Object Lock is enabled on the bucket")
        return True
    except Exception as e:
        error_code = None
        if hasattr(e, 'response') and 'Error' in e.response:
            error_code = e.response['Error'].get('Code')

        if error_code == 'ObjectLockConfigurationNotFoundError':
            logger.error("Object Lock configuration not found on the bucket")
            logger.error("This tool is designed for Object Lock enabled buckets.")
        else:
            logger.error("Failed to check Object Lock configuration: %s", e)

        return False


def cleanup_bucket(config, logger):
    """Main cleanup function"""
    start_time = datetime.datetime.now()
    bucket_name = config['BUCKET_NAME']
    endpoint_url = config['ENDPOINT_URL']
    access_key = config['ACCESS_KEY']
    secret_key = config['SECRET_KEY']

    # Statistics counters
    stats = {
        'total_objects_processed': 0,
        'objects_with_delete_marker_current': 0,
        'objects_fully_deleted': 0,
        'objects_skipped_retention': 0,
        'versions_deleted': 0,
        'delete_markers_deleted': 0,
        'deletion_errors': 0,
        'bulk_delete_batches': 0,
        'total_size_all_objects': 0,
        'total_size_hidden_under_delete_markers': 0,
        'actual_size_cleaned_up': 0
    }

    # Bulk deletion tracking
    delete_batch = []
    batch_number = 1

    try:
        # Setup S3 client
        s3_client_args = {'endpoint_url': endpoint_url}
        if access_key:
            s3_client_args['aws_access_key_id'] = access_key
        if secret_key:
            s3_client_args['aws_secret_access_key'] = secret_key

        s3_client = boto3.client('s3', **s3_client_args)

        # Test bucket access
        try:
            s3_client.head_bucket(Bucket=bucket_name)
            logger.info("Successfully connected to bucket: %s", bucket_name)
        except Exception as e:
            logger.error("Failed to access bucket %s. Error: %s", bucket_name, e)
            return stats

        # Check bucket configuration requirements
        logger.info("Checking bucket configuration...")

        if not check_bucket_versioning(s3_client, bucket_name, logger):
            logger.error("Bucket configuration check failed: Versioning not enabled")
            return stats

        if not check_bucket_object_lock(s3_client, bucket_name, logger):
            logger.error("Bucket configuration check failed: Object Lock not enabled")
            return stats

        logger.info("Bucket configuration validated successfully")

        # List all object versions
        object_versions_map = list_all_object_versions(s3_client, bucket_name, logger)
        total_unique_objects = len(object_versions_map)

        # Calculate size statistics
        for key, versions in object_versions_map.items():
            # Calculate total size of all versions for this object (excluding delete markers)
            object_total_size = sum(v['Size'] for v in versions if not v['IsDeleteMarker'])
            stats['total_size_all_objects'] += object_total_size

            # Check if current version is a delete marker
            current_version = None
            for version in versions:
                if version['IsLatest']:
                    current_version = version
                    break

            if current_version and current_version['IsDeleteMarker']:
                stats['total_size_hidden_under_delete_markers'] += object_total_size

        # Find objects with delete markers as current version
        objects_to_process = []
        for key, versions in object_versions_map.items():
            current_version = None
            for version in versions:
                if version['IsLatest']:
                    current_version = version
                    break

            if current_version and current_version['IsDeleteMarker']:
                objects_to_process.append((key, versions))

        logger.info("Found %d unique objects in bucket", total_unique_objects)
        logger.info("Total size of all objects: %s", human_readable_size(stats['total_size_all_objects']))
        logger.info("Objects with delete marker as current version: %d", len(objects_to_process))
        logger.info("Size hidden under delete markers: %s",
                    human_readable_size(stats['total_size_hidden_under_delete_markers']))
        logger.info("Starting deletion phase...")

        # Process objects in chunks
        objects_processed = 0
        chunk_size = BULK_DELETE_SIZE * 2

        while objects_processed < len(objects_to_process):
            chunk_end = min(objects_processed + chunk_size, len(objects_to_process))
            current_chunk = objects_to_process[objects_processed:chunk_end]

            logger.info("Processing chunk: objects %d-%d of %d",
                        objects_processed + 1, chunk_end, len(objects_to_process))

            # Parallel retention checking
            protected_objects_chunk = set()

            def create_retention_check_args(key_versions_tuple):
                key, versions = key_versions_tuple
                client = create_s3_client_for_retention_check(endpoint_url, access_key, secret_key)
                return (client, bucket_name, key, versions, logger)

            with ThreadPoolExecutor(max_workers=RETENTION_CHECK_THREADS) as executor:
                future_to_key = {
                    executor.submit(check_single_object_retention,
                                    create_retention_check_args(key_versions)): key_versions[0]
                    for key_versions in current_chunk
                }

                for future in as_completed(future_to_key):
                    key = future_to_key[future]
                    try:
                        object_key, has_retention = future.result()
                        if has_retention:
                            protected_objects_chunk.add(object_key)
                    except Exception as e:
                        logger.error("Failed to check retention for object %s: %s", key, e)
                        protected_objects_chunk.add(key)

            # Build delete batch from unprotected objects
            unprotected_objects = []
            for key, versions in current_chunk:
                if key not in protected_objects_chunk:
                    unprotected_objects.append((key, versions))

            logger.info("Found %d unprotected objects in chunk (%d protected)",
                        len(unprotected_objects), len(protected_objects_chunk))

            # Process unprotected objects
            for key, versions in unprotected_objects:
                stats['total_objects_processed'] += 1
                stats['objects_with_delete_marker_current'] += 1

                # Get current version (delete marker) and older versions
                current_version = None
                for version in versions:
                    if version['IsLatest']:
                        current_version = version
                        break

                older_versions = [v for v in versions if not v['IsLatest']]

                # Add versions to delete batch
                versions_to_delete = []
                if not older_versions:
                    versions_to_delete.append(current_version)
                else:
                    versions_to_delete.extend(older_versions)
                    versions_to_delete.append(current_version)

                for version in versions_to_delete:
                    delete_batch.append({
                        'Key': key,
                        'VersionId': version['VersionId'],
                        'IsDeleteMarker': version['IsDeleteMarker'],
                        'Size': version['Size']
                    })

                    # Execute batch when it reaches the configured size
                    if len(delete_batch) >= BULK_DELETE_SIZE:
                        # Save batch to file
                        save_delete_batch_to_file(delete_batch, batch_number, bucket_name, logger)

                        # Execute bulk delete
                        success_count, error_count = execute_bulk_delete(
                            s3_client, bucket_name, delete_batch, logger)

                        stats['bulk_delete_batches'] += 1
                        stats['deletion_errors'] += error_count

                        # Update deletion counters and size
                        if success_count > 0:
                            delete_markers_in_batch = sum(1 for obj in delete_batch
                                                          if obj.get('IsDeleteMarker', False))
                            stats['delete_markers_deleted'] += min(delete_markers_in_batch,
                                                                   success_count)
                            stats['versions_deleted'] += (success_count -
                                                          min(delete_markers_in_batch,
                                                              success_count))

                            # Calculate actual size cleaned up (only for successfully deleted objects)
                            for i, obj in enumerate(delete_batch):
                                if i < success_count:  # Only count successfully deleted objects
                                    stats['actual_size_cleaned_up'] += obj.get('Size', 0)

                        logger.info("Processed batch %d: %d successful, %d errors",
                                    batch_number, success_count, error_count)

                        delete_batch = []
                        batch_number += 1

                if len(delete_batch) >= BULK_DELETE_SIZE:
                    break

            # Update protected objects count
            stats['objects_skipped_retention'] += len(protected_objects_chunk)
            objects_processed = chunk_end

        # Execute remaining batch if any
        if delete_batch:
            save_delete_batch_to_file(delete_batch, batch_number, bucket_name, logger)
            success_count, error_count = execute_bulk_delete(
                s3_client, bucket_name, delete_batch, logger)

            stats['bulk_delete_batches'] += 1
            stats['deletion_errors'] += error_count

            if success_count > 0:
                delete_markers_in_batch = sum(1 for obj in delete_batch
                                              if obj.get('IsDeleteMarker', False))
                stats['delete_markers_deleted'] += min(delete_markers_in_batch,
                                                       success_count)
                stats['versions_deleted'] += (success_count -
                                              min(delete_markers_in_batch,
                                                  success_count))

                # Calculate actual size cleaned up for final batch
                for i, obj in enumerate(delete_batch):
                    if i < success_count:  # Only count successfully deleted objects
                        stats['actual_size_cleaned_up'] += obj.get('Size', 0)

            logger.info("Processed final batch %d: %d successful, %d errors",
                        batch_number, success_count, error_count)

        stats['objects_fully_deleted'] = stats['delete_markers_deleted']

        end_time = datetime.datetime.now()
        duration = end_time - start_time

        # Final statistics
        logger.info("=" * 60)
        logger.info("CLEANUP COMPLETED")
        logger.info("=" * 60)
        logger.info("Duration: %s", duration)
        logger.info("Objects processed: %d", stats['total_objects_processed'])
        logger.info("Objects fully deleted: %d", stats['objects_fully_deleted'])
        logger.info("Objects skipped (protected by retention): %d", stats['objects_skipped_retention'])
        logger.info("Object versions deleted: %d", stats['versions_deleted'])
        logger.info("Delete markers deleted: %d", stats['delete_markers_deleted'])
        logger.info("Deletion errors: %d", stats['deletion_errors'])
        logger.info("Bulk delete batches executed: %d", stats['bulk_delete_batches'])
        logger.info("-" * 60)
        logger.info("SIZE STATISTICS:")
        logger.info("Total size of all objects in bucket: %s",
                    human_readable_size(stats['total_size_all_objects']))
        logger.info("Size hidden under delete markers: %s",
                    human_readable_size(stats['total_size_hidden_under_delete_markers']))
        logger.info("Actual storage space cleaned up: %s",
                    human_readable_size(stats['actual_size_cleaned_up']))
        if stats['total_size_hidden_under_delete_markers'] > 0:
            cleanup_percentage = (stats['actual_size_cleaned_up'] /
                                  float(stats['total_size_hidden_under_delete_markers'])) * 100
            logger.info("Cleanup efficiency: %.1f%% of hidden data was successfully cleaned up",
                        cleanup_percentage)
        logger.info("=" * 60)

        return stats

    except Exception as e:
        logger.error("An unexpected error occurred during cleanup: %s", e)
        return stats


if __name__ == "__main__":
    # --- Main script execution starts here ---
    logger = None  # Initialize logger to None for potential early exit before setup
    try:
        # Load configuration
        config = load_config()

        # Setup logging
        timestamp_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        log_file = "s3_bucket_cleanup_{}.log".format(timestamp_str)
        logger = setup_logger(log_file)

        # Clean up old log files (keep only last 10)
        cleanup_old_logs()

        logger.info("S3 Immutable Bucket Cleanup Tool - Simplified Version")
        logger.info("Bucket name: %s", config['BUCKET_NAME'])
        logger.info("Endpoint URL: %s", config['ENDPOINT_URL'])

        # Run cleanup
        final_stats = cleanup_bucket(config, logger)

        logger.info("Script execution completed.")
    except SystemExit:  # Allow sys.exit() to propagate
        raise
    except Exception as e:
        # Log critical error if logger is available, otherwise print to stderr
        if logger:
            logger.critical(f"Unhandled exception at main execution level: {e}", exc_info=True)
        else:
            print(f"Critical error before logger initialization: {e}", file=sys.stderr)
    sys.exit(1)  # Exit with an error code
